<?php

declare(strict_types=1);

namespace Drupal\data_migration_report\Commands;

use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Database;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Extension\ExtensionPathResolver;
use Drupal\Core\File\FileExists;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGenerator;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\Url;
use Drupal\data_migration_report\Form\CredentialForm;
use Drupal\field\FieldConfigInterface;
use Drupal\migrate_drupal\MigrationConfigurationTrait;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Output\ConsoleOutput;
use Symfony\Component\Yaml\Parser;
use Symfony\Component\Yaml\Yaml;

/**
 * A drush command file.
 */
class DataMigrationReportCommands extends DrushCommands {

  use MigrationConfigurationTrait;
  use StringTranslationTrait;

  /**
   * YAML parser.
   *
   * @var \Symfony\Component\Yaml\Parser
   */
  protected Parser $parser;

  /**
   * The source database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $sourceDatabaseConnection;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected Connection $database;

  /**
   * Path resolver service.
   *
   * @var \Drupal\Core\Extension\ExtensionPathResolver
   */
  protected ExtensionPathResolver $pathResolver;

  /**
   * The state service.
   *
   * @var \Drupal\Core\State\StateInterface
   */
  protected $state;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * The file url generator service.
   *
   * @var \Drupal\Core\File\FileUrlGenerator
   */
  protected FileUrlGenerator $fileUrlGenerator;

  /**
   * The renderer service.
   *
   * @var \Drupal\Core\Render\RendererInterface
   */
  protected RendererInterface $renderer;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected EntityFieldManagerInterface $entityFieldManager;

  /**
   * The source data.
   *
   * @var array
   */
  protected array $source = [];

  /**
   * The destination data.
   *
   * @var array
   */
  protected array $dest = [];

  /**
   * The field and column mapping.
   *
   * @var array
   */
  protected array $fieldColumnMapping = [];

  /**
   * The field preprocess mapping.
   *
   * @var array
   */
  protected array $fieldPreprocessMapping = [];

  /**
   * Constructs a new DataMigrationReportCommands object.
   *
   * @param \Drupal\Core\Database\Connection $database
   *   The database connection.
   * @param \Drupal\Core\Extension\ExtensionPathResolver $path_resolver
   *   Service used to get the path from module machine name.
   * @param \Drupal\Core\State\StateInterface $state
   *   The state service.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system service.
   * @param \Drupal\Core\File\FileUrlGenerator $fileUrlGenerator
   *   The file url generator service.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   *   The entity field manager.
   */
  public function __construct(Connection $database, ExtensionPathResolver $path_resolver, StateInterface $state, FileSystemInterface $file_system, FileUrlGenerator $fileUrlGenerator, RendererInterface $renderer, EntityFieldManagerInterface $entity_field_manager) {
    $this->database           = $database;
    $this->pathResolver       = $path_resolver;
    $this->state              = $state;
    $this->fileSystem         = $file_system;
    $this->fileUrlGenerator   = $fileUrlGenerator;
    $this->renderer           = $renderer;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * Drush Command to run Migration testing.
   *
   * @command generate:content-mapping
   * @usage drush gcm
   * @aliases gcm
   */
  public function generateContentMapping(): void {
    // Check the source database connection and log an error if it's not found.
    $status = $this->checkSourceConnection();

    if (!$status) {
      $this->logger()->error($this->t('No source database connection found.'));
      $this->logger()->error($this->t('Please configure the source database credentials at :config_page', [
        ':config_page' => Url::fromRoute('data_migration_report.migrate_database_credentials', [], ['absolute' => TRUE])->toString(),
      ]));
      return;
    }

    // Generate YAML file for user entity.
    $fields = $this->contentTypeFields('user', 'user');
    $this->generateMappingYaml('user', 'user', $fields);

    // Generate YAML files for node entities.
    $node_types = $this->getValues('SELECT type FROM {node_type}', 'd7', 'one_column');

    if (!empty($node_types)) {
      foreach ($node_types as $type) {
        $fields = $this->contentTypeFields('node', $type);
        $this->generateMappingYaml('node', $type, $fields);
      }
    }

    // Generate YAML files for taxonomy_term entities.
    $taxonomy_types = $this->getValues('SELECT machine_name FROM {taxonomy_vocabulary}', 'd7', 'one_column');

    if (!empty($taxonomy_types)) {
      foreach ($taxonomy_types as $type) {
        $fields = $this->contentTypeFields('taxonomy_term', $type);
        $this->generateMappingYaml('taxonomy_term', $type, $fields);
      }
    }
  }

  /**
   * Returns the fields list of a given entity type bundle.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The entity bundle.
   *
   * @return array
   *   The fields list.
   */
  public function contentTypeFields(string $entity_type, string $bundle): array {
    $fields = [];

    if (!empty($entity_type) && !empty($bundle)) {
      // Query to fetch fields for Drupal 7.
      $d7_fields_query      = "SELECT fc.field_name, fc.type ";
      $d7_fields_query     .= "FROM {field_config} fc ";
      $d7_fields_query     .= "LEFT JOIN {field_config_instance} fci ON fci.field_id = fc.id ";
      $d7_fields_query     .= "WHERE fci.deleted = 0 AND fci.entity_type = :entity_type AND fci.bundle = :bundle";
      $d7_fields_query     .= "ORDER BY fc.field_name ASC";
      $d7_fields_query_args = [
        ':entity_type' => $entity_type,
        ':bundle' => $bundle,
      ];
      $d7_fields            = $this->getValues($d7_fields_query, 'd7', 'two_column', $d7_fields_query_args);

      // Initialize an array for Drupal 10 field information.
      $d10_fields            = [];
      $d10_field_definitions = $this->entityFieldManager->getFieldDefinitions($entity_type, $bundle);

      // Extract field names and types for Drupal 10.
      foreach ($d10_field_definitions as $field_definition) {
        if ($field_definition instanceof FieldConfigInterface) {
          $d10_fields[$field_definition->get('field_name')] = $field_definition->get('field_type');
        }
      }

      // Combine Drupal 7 and Drupal 10 field information into a single array.
      $fields = [
        'd7' => $d7_fields,
        'd10' => $d10_fields,
      ];
    }

    return $fields;
  }

  /**
   * Generates the content mapping yaml file.
   *
   * @param string $entity_type
   *   The entity type.
   * @param string $bundle
   *   The entity bundle.
   * @param array $fields
   *   The fields list.
   */
  public function generateMappingYaml(string $entity_type, string $bundle, array $fields): void {
    // Initialize the YAML content.
    $yaml_content = [
      'entity_type' => $entity_type,
      'bundle' => $bundle,
      'table' => '',
      'fields' => [],
    ];

    // Set the table name based on the entity type.
    $table_map = [
      'user' => 'migrate_map_d7_user',
      'node' => 'migrate_map_d7_node_complete__' . $bundle,
      'taxonomy_term' => 'migrate_map_d7_taxonomy_term__' . $bundle,
    ];
    $yaml_content['table'] = $table_map[$entity_type];

    // Default generated table names, limited to 63 characters.
    $prefix_length = 0;

    if ($database = $this->state->get('migrate_database')) {
      $prefix_length = strlen($database['prefix']);
    }

    $yaml_content['table'] = mb_substr($yaml_content['table'], 0, 63 - $prefix_length);

    // Loop through the fields and add them to the YAML content.
    foreach ($fields['d7'] as $name => $type) {
      $i = count($yaml_content['fields']);

      $yaml_content['fields'][$i]['d7'] = [
        'name' => $name,
        'type' => $type,
      ];

      if (isset($fields['d10'][$name])) {
        $yaml_content['fields'][$i]['d10'] = [
          'name' => $name,
          'type' => $fields['d10'][$name],
        ];
      }
    }

    // Convert the YAML content to YAML format.
    $yaml = Yaml::dump($yaml_content, 4, 2);

    // Prepare the directory and ensure it's writable.
    $directory = 'public://data_migration_report/migration-mapping/';

    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger()->error($this->t('Unable to create directory @directory. The public directory is not writable.', ['@directory' => $directory]));
      return;
    }

    // Save the Migration mapping file.
    $file_path    = $directory . $entity_type . '_' . $bundle . '.yml';
    $mapping_file = $this->fileSystem->saveData($yaml, $file_path, FileExists::Replace);

    if ($mapping_file) {
      $path = $this->fileUrlGenerator->generateAbsoluteString($mapping_file);
      $this->logger()->log('success', $this->t('Migration mapping yml for @entity_type - @bundle generated at @path', [
        '@entity_type' => $entity_type,
        '@bundle' => $bundle,
        '@path' => $path,
      ]));
    }
    else {
      $this->logger()->error($this->t('An error occurred while creating the Migration mapping YAML file.'));
    }
  }

  /**
   * Drush Command to run Migration testing.
   *
   * @param string $entity_type
   *   The type of the content entity being migrated.
   * @param string $bundle
   *   The bundle of the content entity.
   * @param array $options
   *   Array of options.
   *
   * @option integer limit
   *   Number of items to process.
   *
   * @command migration:test
   * @usage drush mt node page --limit=10
   * @aliases mt
   */
  public function migrationTesting(string $entity_type, string $bundle, array $options = ['limit' => 0]): void {
    // Load content data from the YML file.
    $path = 'public://data_migration_report/migration-mapping/';
    $file = $path . $entity_type . '_' . $bundle . '.yml';

    if (!file_exists($file)) {
      $this->logger()->error($this->t('Migration mapping yml file not found at @path', [
        '@path' => $file,
      ]));
      return;
    }

    // Check source database connection and log an error if it's not found.
    $status = $this->checkSourceConnection();

    if (!$status) {
      $this->logger()->error($this->t('No source database connection found.'));
      $this->logger()->error($this->t('Please configure the source database credentials at :config_page', [
        ':config_page' => Url::fromRoute('data_migration_report.migrate_database_credentials', [], ['absolute' => TRUE])->toString(),
      ]));
      return;
    }

    $content_data = $this->getParser()->parse(file_get_contents($file)) ?? [];
    $entity_type  = $content_data['entity_type'];
    $bundle       = $content_data['bundle'];
    $table        = $content_data['table'];

    if (empty($content_data)) {
      $this->logger()->error($this->t('No content found in the YML file.'));
      return;
    }

    // Extract fields data from the content data.
    $errors = $fields = $fields['d7'] = $fields['d10'] = [];

    foreach ($content_data['fields'] as $field) {
      if (isset($field['d7'])) {
        $fields['d7'][] = $field['d7'];
      }
      if (isset($field['d10'])) {
        $fields['d10'][] = $field['d10'];
      }
    }

    // Prepare SQL queries for fetching Drupal 7 IDs.
    $d7_ids_query_args = [];

    if ($entity_type == 'user') {
      $d7_ids_query = "SELECT uid FROM {users} WHERE uid > 0 ORDER BY uid ASC";
    }
    elseif ($entity_type == 'node') {
      $d7_ids_query                 = "SELECT nid FROM {node} WHERE type = :bundle ORDER BY nid ASC";
      $d7_ids_query_args[':bundle'] = $bundle;
    }
    elseif ($entity_type == 'taxonomy_term') {
      $d7_ids_query                 = "SELECT t.tid FROM {taxonomy_term_data} t LEFT JOIN {taxonomy_vocabulary} tv ON tv.vid = t.vid WHERE tv.machine_name = :bundle ORDER BY t.tid ASC";
      $d7_ids_query_args[':bundle'] = $bundle;
    }

    if (!empty($d7_ids_query)) {
      if ($options['limit'] && $options['limit'] > 0) {
        $d7_ids_query = $d7_ids_query . " LIMIT " . $options['limit'];
      }

      $d7_ids = $this->getValues($d7_ids_query, 'd7', 'one_column', $d7_ids_query_args);
    }
    else {
      $d7_ids_query = $this->database->select($content_data['table'], 'n')
        ->fields('n', ['sourceid1'])
        ->orderBy('sourceid1', 'ASC');

      if ($options['limit'] && $options['limit'] > 0) {
        $d7_ids_query->range(0, $options['limit']);
      }

      $d7_ids = $d7_ids_query->execute()->fetchCol();
    }

    if (empty($d7_ids)) {
      $this->logger()->error($this->t('No data found for @entity_type - @bundle in Drupal 7.', [
        '@entity_type' => $entity_type,
        '@bundle' => $bundle,
      ]));
      return;
    }

    // Load destination IDs.
    $d10_ids_query = $this->database->select($content_data['table'], 'n')
      ->fields('n', ['destid1'])
      ->condition('destid1', NULL, 'IS NOT')
      ->orderBy('sourceid1', 'ASC');

    if ($options['limit'] && $options['limit'] > 0) {
      $d10_ids_query->range(0, $options['limit']);
    }

    $d10_ids = $d10_ids_query->execute()->fetchCol();

    // Load content mapping.
    $content_mapping_query = $this->database->select($content_data['table'], 'n')
      ->fields('n', ['sourceid1', 'destid1'])
      ->orderBy('sourceid1', 'ASC');
    $content_mapping = $content_mapping_query->execute()->fetchAllKeyed(0, 1);

    // Load source and destination content.
    $d7_contents  = $this->getEntityValues($entity_type, $content_data['fields'], 'd7', $d7_ids);
    $d10_contents = $this->getEntityValues($entity_type, $content_data['fields'], 'd10', $d10_ids);

    // Preprocess the content.
    $processed_data = $this->executePreprocess($entity_type, $fields, $d7_contents, $d10_contents);
    $d7_contents    = $processed_data['source'];
    $d10_contents   = $processed_data['dest'];

    // Find errors in the content.
    $column_d7  = array_keys(reset($d7_contents));
    $column_d10 = array_keys(reset($d10_contents));

    foreach ($d7_contents as $entity) {
      $installed_drupal_version = intval(\Drupal::VERSION);
      $destination_value        = "d{$installed_drupal_version}_value";

      if ($entity_type == 'user') {
        $id = $content_mapping[$entity['uid']] ?? $entity['uid'];
      }
      elseif ($entity_type == 'node') {
        $id = $content_mapping[$entity['nid']] ?? $entity['nid'];
      }
      elseif ($entity_type == 'taxonomy_term') {
        $id = $content_mapping[$entity['tid']] ?? $entity['tid'];
      }

      if (array_key_exists($id, $d10_contents)) {
        for ($i = 0; $i < count($column_d7); $i++) {
          $d7_column      = $column_d7[$i];
          $d7_field       = $this->fieldColumnMapping['d7'][$d7_column] ?? 'default';
          $d10_column_key = $this->fieldPreprocessMapping[$d7_column] ?? $d7_column;

          if (in_array($d7_column, $column_d10)) {
            $d10_column = $d7_column;

            if (is_array($entity[$d7_column]) && is_array($d10_contents[$id][$d10_column])) {
              if (array_diff($entity[$d7_column], $d10_contents[$id][$d10_column])) {
                $errors[$id][$d7_field][] = [
                  'd7_value' => [
                    $d7_column => $entity[$d7_column],
                  ],
                  $destination_value => [
                    $d10_column_key => $d10_contents[$id][$d10_column],
                  ],
                ];
              }
            }
            else {
              if ($entity[$d7_column] != $d10_contents[$id][$d10_column]) {
                $errors[$id][$d7_field][] = [
                  'd7_value' => [
                    $d7_column => $entity[$d7_column],
                  ],
                  $destination_value => [
                    $d10_column_key => $d10_contents[$id][$d10_column],
                  ],
                ];
              }
            }
          }
          else {
            if (!empty($entity[$d7_column])) {
              $errors[$id][$d7_field][] = [
                'd7_value' => [
                  $d7_column => $entity[$d7_column],
                ],
                $destination_value => [
                  $d10_column_key => '',
                ],
              ];
            }
          }
        }
      }
      else {
        for ($i = 0; $i < count($column_d7); $i++) {
          $d7_column      = $column_d7[$i];
          $d7_field       = $this->fieldColumnMapping['d7'][$d7_column] ?? 'default';
          $d10_column_key = $this->fieldPreprocessMapping[$d7_column] ?? $d7_column;

          if (!empty($entity[$d7_column])) {
            $errors[$id][$d7_field][] = [
              'd7_value' => [
                $d7_column => $entity[$d7_column],
              ],
              $destination_value => [
                $d10_column_key => '',
              ],
            ];
          }
        }
      }
    }

    $total  = count($d7_ids);
    $failed = count($errors);
    $passed = $total - $failed;

    $report_markup = [
      'entity_type' => $entity_type,
      'bundle' => $bundle,
      'table' => $table,
      'total' => $total,
      'passed' => $passed,
      'failed' => $failed,
      'errors' => $errors,
      'pass' => round(($passed / $total) * 100, 2),
      'fail' => round(($failed / $total) * 100, 2),
    ];

    $report_markup = [
      '#theme' => 'migration_report',
      '#data' => $report_markup,
    ];

    $report_markup = $this->renderer->renderInIsolation($report_markup);

    // Print the summary table.
    $output = new ConsoleOutput();
    $output->writeln('<info>Migration Summary</info>');

    if ($entity_type == 'node' || $entity_type == 'taxonomy_term') {
      $type = ucwords(str_replace('_', ' ', $entity_type));
      $summary_header = [$type, 'Total', 'Passed', 'Failed'];
    }
    else {
      $summary_header = ['Type', 'Total', 'Passed', 'Failed'];
    }

    $summary_rows = [
      $bundle => [
        'type' => $bundle,
        'total' => $total,
        'passed' => $passed,
        'failed' => $failed,
      ],
    ];
    $this->printTable($summary_header, $summary_rows, $output);

    // HTML Report Generation.
    $directory = 'public://data_migration_report/migration-reports/';

    // Prepare the directory and ensure it's writable.
    if (!$this->fileSystem->prepareDirectory($directory, FileSystemInterface::CREATE_DIRECTORY)) {
      $this->logger()->error($this->t('Unable to create directory @directory. The public directory is not writable.', ['@directory' => $directory]));
    }

    // Save the Migration Summary Report file.
    $report_file_path = $directory . 'migration_report-' . $entity_type . '-' . $bundle . '.html';
    $report_file      = $this->fileSystem->saveData($report_markup->__toString(), $report_file_path, FileExists::Replace);

    if ($report_file) {
      $path = $this->fileUrlGenerator->generateAbsoluteString($report_file);
      $this->logger()->log('success', $this->t('Migration Report generated at @path', ['@path' => $path]));
    }
    else {
      $this->logger()->error($this->t('An error occurred while creating the Migration Summary Report HTML file.'));
    }
  }

  /**
   * Check the source database connection.
   *
   * @return bool
   *   TRUE if the source database connection is available, FALSE otherwise.
   */
  public function checkSourceConnection(): bool {
    $status = FALSE;

    // Check if the "migrate" connection is available.
    if (Database::getConnection('default', 'migrate')) {
      $status                         = TRUE;
      $this->sourceDatabaseConnection = Database::getConnection('default', 'migrate');
    }
    else {
      // Get the driver and database values from the state.
      $driver   = $this->state->get('migrate_database_driver');
      $database = $this->state->get('migrate_database');

      // If both values are present, create the source database connection.
      if (!empty($driver) && !empty($database)) {
        // Get the namespace and driver from the database types.
        $drivers           = CredentialForm::getDatabaseTypes();
        $reflection        = new \ReflectionClass($drivers[$driver]);
        $install_namespace = $reflection->getNamespaceName();
        // Cut the trailing \Install from namespace.
        $database['namespace']          = substr($install_namespace, 0, strrpos($install_namespace, '\\'));
        $database['driver']             = $driver;
        $this->sourceDatabaseConnection = $this->getConnection($database);
        $status                         = TRUE;
      }
    }

    return $status;
  }

  /**
   * Prints tabular data.
   *
   * @param array $header
   *   The header for the table.
   * @param array $rows
   *   The rows of data to be displayed in the table.
   * @param \Symfony\Component\Console\Output\ConsoleOutput $output
   *   The console output to be used for rendering the table.
   */
  public function printTable(array $header, array $rows, ConsoleOutput $output): void {
    // Initialize table.
    $table = new Table($output);

    // Set the table header.
    $table->setHeaders($header);
    // Add rows to the table.
    $table->addRows($rows);

    // Render the table.
    $table->render();
  }

  /**
   * Get the YAML parser service.
   *
   * @return \Symfony\Component\Yaml\Parser
   *   The YAML parser service.
   */
  protected function getParser(): Parser {
    // Create a new YAML parser if it has not been created yet.
    if (!isset($this->parser)) {
      $this->parser = new Parser();
    }

    // Return the YAML parser service.
    return $this->parser;
  }

  /**
   * Get the values of the entity fields from the database.
   *
   * @param string $entity_type
   *   The type of entity.
   * @param array $fields
   *   The fields of the entity.
   * @param string $drupal_version
   *   The version of Drupal.
   * @param array $ids
   *   The IDs of the entities.
   *
   * @return array
   *   The values of the entity fields.
   */
  public function getEntityValues(string $entity_type, array $fields, string $drupal_version, array $ids): array {
    // Initialize the base columns array and the query.
    $base_columns = [];
    $entity_table = $query = '';

    // Set prefix and entity table based on the entity type and Drupal version.
    if ($entity_type == 'user') {
      $prefix       = $drupal_version == 'd10' ? 'user__' : 'field_data_';
      $entity_table = $drupal_version == 'd10' ? '{users_field_data}' : '{users}';
      $base_columns = [
        $entity_table . '.uid',
        $entity_table . '.name',
        $entity_table . '.mail',
        $entity_table . '.status',
        $entity_table . '.timezone',
      ];
    }
    elseif ($entity_type == 'node') {
      $prefix         = $drupal_version == 'd10' ? 'node__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{node_field_data}' : '{node}';
      $base_columns   = [
        $entity_table . '.nid',
        $entity_table . '.title',
        $entity_table . '.type',
        $entity_table . '.status',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.langcode' : $entity_table . '.language';
    }
    elseif ($entity_type == 'taxonomy_term') {
      $prefix         = $drupal_version == 'd10' ? 'taxonomy_term__' : 'field_data_';
      $entity_table   = $drupal_version == 'd10' ? '{taxonomy_term_field_data}' : '{taxonomy_term_data}';
      $base_columns   = [
        $entity_table . '.tid',
        $entity_table . '.vid',
        $entity_table . '.name',
      ];
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.description__value' : $entity_table . '.description';
      $base_columns[] = $drupal_version == 'd10' ? $entity_table . '.description__format' : $entity_table . '.format';
    }

    // Join the base columns with the field columns.
    $columns = implode(",", $base_columns);
    $tables  = $entity_table;

    // Join tables with the field tables and add the field columns to the query.
    foreach ($fields as $field) {
      if (isset($field[$drupal_version]['name']) && $field[$drupal_version]['type']) {
        $field_name    = $field[$drupal_version]['name'];
        $field_type    = $field[$drupal_version]['type'];
        $field_columns = $this->getFieldColumnList($field_name, $field_type, $prefix, $drupal_version);
        $columns      .= $field_columns;

        if ($entity_type == 'user') {
          $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.uid";
        }
        elseif ($entity_type == 'node') {
          $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.nid";
        }
        elseif ($entity_type == 'taxonomy_term') {
          $tables .= " LEFT JOIN {$prefix}{$field_name} ON {$prefix}{$field_name}.entity_id = {$entity_table}.tid";
        }
      }
    }

    // Build the query and execute it to get the values.
    if ($entity_type == 'user') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.uid IN (:ids[])";
    }
    elseif ($entity_type == 'node') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.nid IN (:ids[])";
    }
    elseif ($entity_type == 'taxonomy_term') {
      $query = "SELECT $columns FROM $tables WHERE {$entity_table}.tid IN (:ids[])";
    }

    $query_args = [
      ':ids[]' => $ids,
    ];

    return $this->getValues($query, $drupal_version, $entity_type, $query_args);
  }

  /**
   * Get the field columns for a given field name and field type.
   *
   * @param string $field_name
   *   The name of the field.
   * @param string $field_type
   *   The type of the field.
   * @param string $prefix
   *   The prefix to use for the field tables.
   * @param string $drupal_version
   *   The Drupal version.
   *
   * @return string
   *   The list of field columns.
   */
  public function getFieldColumnList(string $field_name, string $field_type, string $prefix, string $drupal_version): string {
    // Initialize variables.
    $columns       = [];
    $field_columns = "";

    // Get the field type mapping.
    $path               = $this->pathResolver->getPath('module', 'data_migration_report');
    $file               = $path . '/field_types.yml';
    $field_type_data    = $this->getParser()->parse(file_get_contents($file)) ?? [];
    $field_type_mapping = $field_type_data[$drupal_version][$field_type] ?? [];

    foreach ($field_type_mapping as $column) {
      // Generate the field columns.
      $column    = str_replace("{field_name}", $field_name, $column);
      $columns[] = "{$prefix}{$field_name}.$column";

      // Store the field column mapping.
      $this->fieldColumnMapping[$drupal_version][$column] = $field_name;
    }

    // Combine the field columns into a string.
    if (!empty($columns)) {
      $field_columns = "," . implode(",", $columns);
    }

    return $field_columns;
  }

  /**
   * Retrieve values from the database based on the given query.
   *
   * @param string $query
   *   The SQL query to execute.
   * @param string $drupal_version
   *   The Drupal version.
   * @param string $result_type
   *   The type of result to fetch.
   * @param array $arguments
   *   The arguments to pass to the query.
   *
   * @return array
   *   An associative array containing the fetched values.
   */
  public function getValues(string $query, string $drupal_version, string $result_type, array $arguments = []): array {
    // Determine the database connection based on the Drupal version.
    if ($drupal_version == 'd10') {
      $connection = $this->database;
    }

    if ($drupal_version == 'd7') {
      $connection = $this->sourceDatabaseConnection;
    }

    // If a database connection exists, execute the query and fetch the values.
    if (!empty($connection)) {
      $result = $connection->query($query, $arguments);

      if ($result_type == 'user') {
        return $result->fetchAllAssoc('uid', \PDO::FETCH_ASSOC);
      }
      elseif ($result_type == 'node') {
        return $result->fetchAllAssoc('nid', \PDO::FETCH_ASSOC);
      }
      elseif ($result_type == 'taxonomy_term') {
        return $result->fetchAllAssoc('tid', \PDO::FETCH_ASSOC);
      }
      elseif ($result_type == 'one_column') {
        return $result->fetchAllKeyed(0, 0);
      }
      elseif ($result_type == 'two_column') {
        return $result->fetchAllKeyed(0, 1);
      }
    }

    // If no values were fetched, return an empty array.
    return [];
  }

  /**
   * Preprocess the data before comparing.
   *
   * @param string $entity_type
   *   The entity type.
   * @param array $fields
   *   An array of fields.
   * @param array $source
   *   The source data.
   * @param array $dest
   *   The destination data.
   *
   * @return array
   *   An array containing the preprocessed source and destination data.
   */
  public function executePreprocess(string $entity_type, array $fields, array $source, array $dest): array {
    // Set the source and destination data.
    $this->source = $source;
    $this->dest   = $dest;

    foreach ($fields as $data) {
      foreach ($data as $field) {
        // Preprocessors field types.
        $method = str_replace("_", "", lcfirst(ucwords($field['type'], " /_")));
        if (method_exists(self::class, $method)) {
          call_user_func([self::class, $method], $field);
        }

        // Preprocessors field specific.
        $method = str_replace("_", "", lcfirst(ucwords($field['name'], " /_")));
        if (method_exists(self::class, $method)) {
          call_user_func([self::class, $method], $field);
        }
      }
    }

    // Preprocessors based on the main entity type.
    $method = str_replace("_", "", lcfirst(ucwords($entity_type, " /_")));
    if (method_exists(self::class, $method)) {
      call_user_func([self::class, $method], NULL);
    }

    // Return the preprocessed source and destination data.
    return ['source' => $this->source, 'dest' => $this->dest];
  }

  /**
   * Preprocessor for email field type.
   *
   * @param array $field
   *   The field definition containing the name of the email field.
   */
  public function email(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate over each record in the destination data.
    foreach ($this->dest as $key => $value) {
      // If the record has a field named {field_name}_value, rename it to
      // {field_name}_email and remove the {field_name}_value field.
      if (array_key_exists($field_name . '_value', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_email'] = $this->dest[$key][$field_name . '_value'];
        unset($this->dest[$key][$field_name . '_value']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_email'] = $field_name . '_value';
      }
    }
  }

  /**
   * Preprocessor for text_with_summary field type.
   *
   * This function removes line breaks and tabs from the source and destination
   * values of the field.
   *
   * @param array $field
   *   The field definition.
   */
  public function textWithSummary(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Remove line breaks and tabs from the source value.
    foreach ($this->source as $key => $value) {
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        $this->source[$key][$field_name . '_value'] = preg_replace('/[\r\n\t]/', '', $value[$field_name . '_value']);
      }
    }

    // Remove line breaks and tabs from the destination value.
    foreach ($this->dest as $key => $value) {
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        $this->dest[$key][$field_name . '_value'] = preg_replace('/[\r\n\t]/', '', $value[$field_name . '_value']);
      }
    }
  }

  /**
   * Preprocessor for datetime field type.
   *
   * @param array $field
   *   The field information.
   */
  public function datetime(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate over the destination array.
    foreach ($this->dest as $key => $value) {
      // Check if the field value exists and is not empty.
      if (array_key_exists($field_name . '_value', $value) && !empty($value[$field_name . '_value'])) {
        // Check if the datetime value has the "T" character.
        if (str_contains($value[$field_name . '_value'], "T")) {
          // Replace the "T" character with a space.
          $this->dest[$key][$field_name . '_value'] = str_replace('T', ' ', $value[$field_name . '_value']);
        }

        // Check if the datetime value has the ':' character.
        if (!strpos($value[$field_name . '_value'], ":")) {
          // Add ' 00:00:00' to the datetime value.
          $this->dest[$key][$field_name . '_value'] = $value[$field_name . '_value'] . ' 00:00:00';
        }
      }
    }
  }

  /**
   * Preprocessor for user_reference field type.
   *
   * This function is a preprocessor for the user_reference field type. It
   * preprocesses the user reference field by calling the
   * DataMigrationReportCommands::preprocessEntityReference() method with the
   * field and '_uid' as arguments.
   *
   * @param array $field
   *   The user reference field.
   */
  public function userReference(array $field): void {
    $this->preprocessEntityReference($field, '_uid');
  }

  /**
   * Preprocessor for node_reference field type.
   *
   * This function is a preprocessor for the node_reference field type. It calls
   * the preprocessEntityReference() function with the field and '_nid' as
   * arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function nodeReference(array $field): void {
    $this->preprocessEntityReference($field, '_nid');
  }

  /**
   * Preprocessor for taxonomy_term_reference field type.
   *
   * This function is a preprocessor for the taxonomy_term_reference field type.
   * It calls the preprocessEntityReference() function with the field and '_tid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function taxonomyTermReference(array $field): void {
    $this->preprocessEntityReference($field, '_tid');
  }

  /**
   * Preprocessor for file field type.
   *
   * This function is a preprocessor for the file field type.
   * It calls the preprocessEntityReference() function with the field and '_fid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function file(array $field): void {
    $this->preprocessEntityReference($field, '_fid');
  }

  /**
   * Preprocessor for image field type.
   *
   * This function is a preprocessor for the image field type.
   * It calls the preprocessEntityReference() function with the field and '_fid'
   * as arguments.
   *
   * @param array $field
   *   The field being processed.
   */
  public function image(array $field): void {
    $this->preprocessEntityReference($field, '_fid');
  }

  /**
   * Preprocessor for entity reference field types.
   *
   * This function is used to preprocess entity reference fields in the
   * destination data. It changes the field name to match the field name in
   * Drupal 10 and removes the '_target_id' suffix.
   *
   * @param array $field
   *   The field definition array.
   * @param string $column_suffix
   *   The suffix to be appended to the field name.
   */
  public function preprocessEntityReference(array $field, string $column_suffix): void {
    // Get the field name.
    $field_name = $field['name'];

    // Iterate through the destination data array.
    foreach ($this->dest as $key => $value) {
      // Check if the field name with '_target_id' suffix exists in the
      // destination data.
      if (array_key_exists($field_name . '_target_id', $this->dest[$key])) {
        // Change the field name to match the field name in Drupal 7.
        $this->dest[$key][$field_name . $column_suffix] = $this->dest[$key][$field_name . '_target_id'];
        unset($this->dest[$key][$field_name . '_target_id']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . $column_suffix] = $field_name . '_target_id';
      }
    }
  }

  /**
   * Preprocessor for link field type.
   *
   * This function preprocesses link field type by renaming '_uri' to '_url'
   * and '_options' to '_attributes' in the destination array.
   *
   * @param array $field
   *   The field array.
   */
  public function link(array $field): void {
    // Get the field name.
    $field_name = $field['name'];

    // Loop through the destination array.
    foreach ($this->dest as $key => $value) {
      // Rename '_uri' to '_url'.
      if (array_key_exists($field_name . '_uri', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_url'] = $this->dest[$key][$field_name . '_uri'];
        unset($this->dest[$key][$field_name . '_uri']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_url'] = $field_name . '_uri';
      }

      // Rename '_options' to '_attributes'.
      if (array_key_exists($field_name . '_options', $this->dest[$key])) {
        $this->dest[$key][$field_name . '_attributes'] = $this->dest[$key][$field_name . '_options'];
        unset($this->dest[$key][$field_name . '_options']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping[$field_name . '_attributes'] = $field_name . '_options';
      }
    }
  }

  /**
   * Preprocessor for node entity type.
   *
   * This preprocessor renames the 'langcode' field to 'language' in the
   * destination array. It also updates the fieldPreprocessMapping array to
   * reflect the change.
   */
  public function node(): void {
    foreach ($this->dest as $key => $value) {
      if (array_key_exists('langcode', $this->dest[$key])) {
        $this->dest[$key]['language'] = $this->dest[$key]['langcode'];
        unset($this->dest[$key]['langcode']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['language'] = 'langcode';
      }
    }
  }

  /**
   * Preprocessor for taxonomy_term entity type.
   *
   * This function maps the vocabulary ID from the source to the destination
   * if it exists in the mapping table. It also renames the 'description'
   * field to 'description__value' and the 'description__format' field to
   * 'format'.
   */
  public function taxonomyTerm(): void {
    // Fetch the mapping of source vocabulary IDs to destination vocabulary IDs.
    $taxonomy_vocabulary_mapping = $this->database->select('migrate_map_d7_taxonomy_vocabulary', 'mtv')
      ->fields('mtv', ['destid1', 'sourceid1'])
      ->execute()->fetchAllKeyed(0, 1);

    // If 'description' field is empty, clear the 'format' field if it exists.
    foreach ($this->source as $key => $value) {
      if (array_key_exists('description', $this->source[$key]) && empty($this->source[$key]['description'])) {
        if (array_key_exists('format', $this->source[$key]) && !empty($this->source[$key]['format'])) {
          $this->source[$key]['format'] = '';
        }
      }
    }

    // Map the vocabulary ID from the source to the destination if it exists.
    foreach ($this->dest as $key => $value) {
      if (array_key_exists('vid', $this->dest[$key])) {
        if (!empty($this->dest[$key]['vid']) && array_key_exists($this->dest[$key]['vid'], $taxonomy_vocabulary_mapping)) {
          $this->dest[$key]['vid'] = $taxonomy_vocabulary_mapping[$this->dest[$key]['vid']];
        }
      }

      // Rename the 'description' field to 'description__value'.
      if (array_key_exists('description__value', $this->dest[$key])) {
        $this->dest[$key]['description'] = $this->dest[$key]['description__value'];
        unset($this->dest[$key]['description__value']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['description'] = 'description__value';
      }

      // Rename the 'description__format' field to 'format'.
      if (array_key_exists('description__format', $this->dest[$key])) {
        $this->dest[$key]['format'] = $this->dest[$key]['description__format'];
        unset($this->dest[$key]['description__format']);

        // Keep track of the field renaming for later use.
        $this->fieldPreprocessMapping['format'] = 'description__format';
      }
    }
  }

}
